/* ----------------------------------------------------------------
 *
 * casca.c
 *
 * CASCA (Campanas All Sky Camera) Scheduler
 *
 * Christoph Birk (birk@carnegiescience.edu)
 *
 * ---------------------------------------------------------------- */

#define DEBUG           1

#define PROJECT_ID      23

/* ---------------------------------------------------------------- */

#define _REENTRANT

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#if defined(WRITE_JPG) || defined(WRITE_JPEG)
#include <jpeglib.h>
#endif
#if defined(WRITE_PNG)
#include <png.h>
#endif

#define IS_CASCA_C

#include "casca.h"
#include "utils.h"
#include "tcpip.h"
#include "zwo.h"
#include "fits.h"

/* ---------------------------------------------------------------- */

#define PREFUN          __func__

/* ---------------------------------------------------------------- */

enum casca_states_enum { STATE_IDLE=0,
                         STATE_NIGHT,
                         STATE_DAY };

static char   configfile[512];
#ifdef SIM_ONLY
static int    offtime=0,sim_delta=0,speedup=0;
#else
static const int offtime=0;
#endif
static double sitelong,sitelat,siteelev;
static int    state=STATE_IDLE;
static volatile int done=0,suspend=0;

static ZwoStruct *zwoCam=NULL,*dayCam=NULL;

static int    ccd_gain,ccd_offset,ccd_setpoint,binning;
static char   datadir[512];
static char   site[32],detector[32],window[64];
static const char *instrument="CASCA";
static double pixsize;
static int    nfilters=0;
static Observation *observations=NULL;

static int    day_gain,day_offset;
static float  day_exptime;

static time_t last_query=0,last_loop=0;
static int    period=60;

#ifdef SIM_ONLY
#define DEF_ZWOHOST  "nuc03"           /* nuc03 */
#define DEF_DAYHOST  ""                /* rock64 */
#else
#define DEF_ZWOHOST  ""
#define DEF_DAYHOST  ""
#endif

/* ---------------------------------------------------------------- */

static char* prtime(time_t ut,char *buf)
{
  struct tm res;
  if (ut <= 0) ut = cor_time(0);
  strftime(buf,64,"%FT%H:%M:%S",gmtime_r(&ut,&res));
  return buf;
}

/* ---------------------------------------------------------------- */

void logit(time_t ut,const char* who,const char* what,int keep)
{
  char buf[32];
  static FILE* fp=NULL;

  if (ut <= 0) ut = cor_time(offtime);
  if (!fp) { char file[512]; struct tm res;
    strftime(buf,sizeof(buf),"%Y_%W",localtime_r(&ut,&res));
    sprintf(file,"%s/casca%s.log","/var/tmp",buf); // ?Povilas
    fp = fopen(file,"a");
    if (!fp) fprintf(stderr,"failed to open logfile '%s'\n",file);
  }
  if (fp) { 
#if (DEBUG > 0)
    fprintf(stdout,"%s: %s - %s\n",get_ut_datestr(buf,ut),who,what);
#endif
    fprintf(fp,    "%s: %s - %s\n",get_ut_datestr(buf,ut),who,what);
    if      (keep == 0) { fclose(fp); fp=NULL; }
    else if (keep >  1) { fflush(fp); }
  }
}

/* ---------------------------------------------------------------- */

static int handle_tcpip_request(char* command,char* answer)
{
  const char* SEP=" \t\r\n";
  char buffer[1024],cmd[128]="",par[128]="",*p,*saveptr;

  msleep(50);                          /* throttle */
  strcpy(answer,"-Etoolong\n");
  if (strlen(command) >= sizeof(cmd)) return 1; /* terminate server */

  strcpy(buffer,command);              /* remove SPC,TAB,CR,LF */
  p = strtok_r(buffer,SEP,&saveptr); if (p) strcpy(cmd,p);
  if ((p=strtok_r(NULL,SEP,&saveptr)) != NULL) strcpy(par,p);
  while ((p=strtok_r(NULL,SEP,&saveptr)) != NULL) {
    strcat(par," "); strcat(par,p);
  }
  sprintf(buffer,"TCP/IP<--: %s %s",cmd,par);
  logit(0,PREFUN,buffer,2);

  strcpy(answer,"ok");
  if        (!strcasecmp(cmd,"version")) {
    sprintf(answer,"%s",P_VERSION);
  } else if (!strcasecmp(cmd,"suspend")) {  /* v0011 todo */
    suspend = 1; // shutdown ?Povilas
  } else if (!strcasecmp(cmd,"resume")) { 
    suspend = 0;
  } else if (!strcasecmp(cmd,"quit")) {
    done = 1;
  } else {
    sprintf(answer,"-Eunknown_command '%s'",cmd);
  }

  sprintf(buffer,"TCP/IP-->: %s",answer);
  logit(0,PREFUN,buffer,2);
  strcat(answer,"\n");

  return 0;
}

/* ---------------------------------------------------------------- */

static double get_el(time_t ut,int obj)
{
  double ra,dec,az,el;

  if (obj==1) get_moon(ut,sitelong,sitelat,&ra,&dec,&el,&az,NULL);
  else        get_sun (ut,sitelong,sitelat,&ra,&dec,&el,&az);
  return el;
}

/* --- */

static void find_object(time_t ut,int obj,int *up,time_t *rise,time_t *set)
{
  double el = get_el(ut,obj);
  *up = (el < 0) ? 0 : 1;
  if (*up) {                           /* object is up */
    time_t t = ut;                     /* rise is in the past */
    do { t -= 1+(int)(el*el); el = get_el(t,obj); } while (el > 0);
    *rise = t+1;
    assert(*rise <= ut);
    t = ut;                            /* set is in the future */
    do { t += 1+(int)(el*el); el = get_el(t,obj); } while (el > 0);
    *set  = t;
    assert(*set  > ut);
  } else {                             /* object is down */
    time_t t = ut;                     /* rise is in the future */
    do { t += 1+(int)(el*el); el = get_el(t,obj); } while (el < 0);
    *rise = t;
    assert(*rise > ut);
    t = ut;                            /* set is in the past */
    do { t -= 1+(int)(el*el); el = get_el(t,obj); } while (el < 0);
    *set  = t+1;
    assert(*set  <= ut);
  }
  // printf("obj=%d up=%d (%ld,%ld,%ld)\n",obj,*up,*rise,ut,*set);
}

/* ---------------------------------------------------------------- */

static int setup_camera(ZwoStruct* zwo,const char* win,int bits)
{
  int  value;
  char buf[256];

  assert(zwo->sock < 0);
  int err = zwo_open(zwo);
  if (err) {
    sprintf(buf,"failed to connect to camera (%s:%d), err=%d",
            zwo->host,zwo->port,err);
    logit(0,PREFUN,buf,2);
    return err;
  }

  if (!err) { int x,y,w,h;
    sprintf(buf,"serverVersion= %s",zwo->serverVersion); 
    logit(0,PREFUN,buf,1);
    sprintf(buf,"geometry= %d %d %d %d",zwo->maxWidth,
           zwo->maxHeight,zwo->hasCooler,zwo->isColor);
    logit(0,PREFUN,buf,1);
    if (*win) { int x1,x2,y1,y2;
      sscanf(win,"[%d:%d,%d:%d]",&x1,&x2,&y1,&y2);
      x = x1; y = y1; w = x2-x1+1; h = y2-y1+1;
      while (w % 8) w--;
      while (h % 2) h--;
    } else {
      x = 0; y = 0; w = zwo->maxWidth/binning; h = zwo->maxHeight/binning;
      assert((w % 8) == 0);
      assert((h % 2) == 0);
    }
    err = zwo_setup(zwo,x,y,w,h,binning,bits);
    sprintf(buf,"aoiW=%d, H=%d",zwo->aoiW,zwo->aoiH);
    logit(0,PREFUN,buf,1);
  }

  if (!err && zwo->hasCooler) {
    err = zwo_tempcon(zwo,ccd_setpoint); // TODO day_ ?Povilas
  }
  value = (zwo == zwoCam) ? ccd_gain : day_gain;
  if (!err) err = zwo_gain(zwo,value);

  value = (zwo == zwoCam) ? ccd_offset : day_offset;
  if (!err) err = zwo_offset(zwo,value);

  if (err) {
    sprintf(buf,"failed to setup camera (err=%d)",err);
    logit(0,PREFUN,buf,2);
    zwo_close(zwo);
  }
  return err;
}

/* ---------------------------------------------------------------- */

static int take_nightime_exposure(Observation* obs)
{
  int    up,err=0;
  time_t rise,set;
  double ra,dec,moon_el,moon_az,moon_ph,sun_az,sun_el;
  char   buf[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,obs->name);
#endif

  time_t ut = cor_time(offtime);
  int npix = zwoCam->aoiW * zwoCam->aoiH;
  assert(zwoCam->bits == 16);
  u_short *data = (void*)malloc(npix*sizeof(u_short));
  get_moon(ut,sitelong,sitelat,&ra,&dec,&moon_el,&moon_az,&moon_ph);
  int moonup = (moon_el < 0) ? 0 : 1;
  float exptime = (moonup) ? obs->time_up : obs->time_dn;
  double illum = 100.0*pow(sin(moon_ph/(2.0*RADS)),2.0);
  if (!err) err = zwo_exptime(zwoCam,exptime);
  if (!err) err = zwo_expose(zwoCam);
  if (!err) { double qt=0;
    if (ut-last_query >= period) {     /* once per sequence */
      qt = tele_query();               /* update telescope pointings */
      last_query = ut;
    }
    double dt = 1.0+exptime-qt;
    if (dt > 0) msleep((int)(1000.0*dt));
    err = zwo_data(zwoCam,data,2);
  }
  if (!err) { char file[512]; struct tm res;
    FITS* f = fits_create(zwoCam->aoiW,zwoCam->aoiH,zwoCam->bits);
    strftime(buf,sizeof(buf),"%Y%m%dut%H%M%S",gmtime_r(&ut,&res));
    sprintf(file,"%s/%s_%s%s.fits",datadir,site,obs->name,buf);
    if (fits_open(f,file) == 0) {
      logit(0,"write FITS",file,2);
      fits_char (f,"DATE-OBS",prtime(ut,buf),NULL);
      fits_char (f,"DETECTOR",detector,NULL);
      fits_char (f,"INSTRUME",instrument,NULL);
      fits_float(f,"SITELONG",sitelong,5,NULL);
      fits_float(f,"SITELAT",sitelat ,5,NULL);
      fits_float(f,"SITEELEV",siteelev,0,NULL);
      fits_float(f,"PIXSIZE",pixsize,3,NULL);
      fits_float(f,"EXPTIME",exptime,4,NULL);
      fits_char (f,"FILTER",obs->name,NULL);
      fits_char (f,"WINDOW",window,NULL);
      fits_int  (f,"BINNING",binning,"CCD binning");
      fits_int  (f,"CCD-GAIN",ccd_gain,"CCD gain"); /* v0010 */
      fits_int  (f,"CCD-OFFS",ccd_offset,"CCD offset");
      fits_float(f,"CCD-TEMP",zwoCam->tempSensor,1,"CCD-temperature [C]");
      fits_float(f,"CCD-COOL",zwoCam->percent,0,"CCD-cooler [%]");
      find_object(ut,1,&up,&rise,&set); 
      strcpy(buf,(up) ? "MOONUP" : "MOONDOWN"); 
      fits_char (f,"MOONSTAT",buf,"state of the moon");
      fits_char (f,"MOONRISE",prtime(rise,buf),"moon rise");
      fits_char (f,"MOONSET",prtime(set,buf),"moon set");
      fits_float(f,"MOONPHZ",illum,2,"moon illumination [%]");
      fits_float(f,"MOONAZ",moon_az,4,"moon azimuth");
      fits_float(f,"MOONEL",moon_el,4,"moon elevation");
      find_object(ut,0,&up,&rise,&set);
      fits_char (f,"SUNRISE",prtime(rise,buf),"sun rise");
      fits_char (f,"SUNSET",prtime(set,buf),"sun set");
      get_sun(ut,sitelong,sitelat,&ra,&dec,&sun_el,&sun_az);
      fits_float(f,"SUNAZ",sun_az,4,"sun azimuth");
      fits_float(f,"SUNEL",sun_el,4,"sun elevation");
      if (*baade_host) {
        fits_float(f,"BAADE_AZ",baade_az, 4,"Baade azimuth");
        fits_float(f,"BAADE_EL",baade_el, 4,"Baade elevation");
      }
      if (*clay_host) {
        fits_float(f,"CLAY_AZ", clay_az,  4,"Clay azimuth");
        fits_float(f,"CLAY_EL", clay_el,  4,"Clay elevation");
      }
      if (*dupont_host) {
        fits_float(f,"C100_AZ", dupont_az,4,"DuPont azimuth");
        fits_float(f,"C100_EL", dupont_el,4,"DuPont elevation");
      }
      if (*dimm_host) {
        fits_float(f,"DIMM_AZ", dimm_az,  4,"DIMM azimuth");
        fits_float(f,"DIMM_EL", dimm_el,  4,"DIMM elevation");
      }
      if (*baade_host || *clay_host) {
        fits_float(f,"ENVTEM",env_temp,2,"ENV temperature");
        fits_float(f,"ENVPRE",env_pres,2,"ENV pressure");
        fits_float(f,"ENVHUM",env_humi,2,"ENV humidity");
        fits_float(f,"ENVWIN",env_wind,2,"ENV wind speed");
        fits_float(f,"ENVDIR",env_wdir,2,"ENV wind direction");
      }
      fits_char (f,"SOFTWARE",P_VERSION,"software version");
      fits_endHeader(f);
      fits_writeData(f,data);
      fits_close(f);
    } else { char buffer[1024];
      sprintf(buffer,"failed to open file '%s'",file);
      logit(0,PREFUN,buffer,2);
    }
    fits_free(f); 
  } /* endif(err) -- reported by caller */

  free((void*)data);

  return err;
}

/* ---------------------------------------------------------------- */

#if defined(WRITE_JPG) || defined(WRITE_JPEG)
static int  write_jpg(const char* file,u_char* data,int w,int h,int q)
{
  struct jpeg_compress_struct cinfo; 
  struct jpeg_error_mgr jerr;

  FILE *fp = fopen(file,"wb"); if (!fp) return -1;
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);
  jpeg_stdio_dest(&cinfo,fp);
  cinfo.image_width = w;
  cinfo.image_height = h;
  cinfo.input_components = 3;
  cinfo.in_color_space = JCS_RGB;
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo,q,TRUE);  // ?Povilas
  jpeg_start_compress(&cinfo,TRUE);
  int row_stride = w * 3;
  JSAMPROW row_pointer[1];
  while (cinfo.next_scanline < cinfo.image_height) {
    row_pointer[0] = (u_char*)data + cinfo.next_scanline * row_stride;
    (void)jpeg_write_scanlines(&cinfo,row_pointer,1);
  }
  jpeg_finish_compress(&cinfo);        /* 0.2 sec [5496 3672] */
  fclose(fp);
  jpeg_destroy_compress(&cinfo); // todo add exif meta data
 
  return 0;
}
#endif

/* --- */

#if defined(WRITE_PNG)
static int  write_png(const char* file,u_char* data,int w,int h,int q)
{
  int i;
  png_structp png_ptr;
  png_infop info_ptr;

  FILE *fp = fopen(file,"wb"); if (!fp) return -1;

  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  info_ptr = png_create_info_struct(png_ptr);
  png_init_io(png_ptr,fp);
  png_set_IHDR(png_ptr, info_ptr, w, h, 8,
                     PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
                     PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
  if (q >= 0) png_set_compression_level(png_ptr,q); /* default=6 NEW v0013 */
  png_write_info(png_ptr, info_ptr);
  png_bytep *row_pointers = (u_char**)malloc(h*sizeof(u_char*));
  for (i=0; i<h; i++) row_pointers[i] = data + i*3*w;
  png_write_image(png_ptr, row_pointers);
  png_write_end(png_ptr, NULL);

  fclose(fp);
  free(row_pointers);
  png_destroy_write_struct(&png_ptr, &info_ptr);

  return 0;
}
#endif

/* --- */

#if defined(WRITE_JPG) || defined(WRITE_JPEG) || defined(WRITE_PNG)
static void write_meta(const char* file,time_t ut,float exptime) 
{
  time_t rise,set;
  double ra,dec,az,el;
  char   name[512],buf[128],*p;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,file);
#endif

  assert(strlen(file) < sizeof(name));
  strcpy(name,file); p = strchr(name,'.'); if (!p) return;
  *p = '\0'; strcat(p,".txt");
  FILE* fp = fopen(name,"w"); if (!fp) return;
  if (fp) { int up;
    fprintf(fp,"INSTRUME= '%s'\n",instrument);
    fprintf(fp,"SITELONG= %.5f\n",sitelong);
    fprintf(fp,"SITELAT=  %.5f\n",sitelat);
    fprintf(fp,"SITEELEV= %.0f\n",siteelev);
    fprintf(fp,"PIXSIZE=  %.3f\n",pixsize);
    fprintf(fp,"EXPTIME=  %.4f\n",exptime);
    fprintf(fp,"WINDOW=   '%s'\n",window);
    fprintf(fp,"BINNING=  %d\n",binning);
    fprintf(fp,"CCD-GAIN= %d\n",day_gain);
    fprintf(fp,"CCD-OFFS= %d\n",day_offset);
    find_object(ut,0,&up,&rise,&set);
    fprintf(fp,"SUNRISE=  '%s'\n",prtime(rise,buf));
    fprintf(fp,"SUNSET =  '%s'\n",prtime(set,buf));
    get_sun(ut,sitelong,sitelat,&ra,&dec,&el,&az);
    fprintf(fp,"SUNAZ=    %.4f\n",az);
    fprintf(fp,"SUNEL=    %.4f\n",el);
    if (*baade_host || *clay_host) {
      fprintf(fp,"ENVTEM=   %.2f\n",env_temp);
      fprintf(fp,"ENVPRE=   %.2f\n",env_pres);
      fprintf(fp,"ENVHUM=   %.2f\n",env_humi);
      fprintf(fp,"ENVWIN=   %.2f\n",env_wind);
      fprintf(fp,"ENVDIR=   %.2f\n",env_wdir);
    }
    fprintf(fp,"SOFTWARE= '%s'\n",P_VERSION);
    fclose(fp);
  }
}
#endif

/* --- */

static int take_daytime_exposure(time_t ut)
{
  int    up;
  time_t rise,set;
  double ra,dec,el,az;
  char   tst[64],buf[512];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%ld)\n",PREFUN,ut);
#endif

  int npix = dayCam->aoiW * dayCam->aoiH;
  void* data = (void*)malloc(npix*dayCam->bits/8);
  float exptime = day_exptime; // todo depends ?Povilas
  int  err = zwo_exptime(dayCam,exptime);
  if (!err) err = zwo_expose(dayCam);
  if (!err) {
    double qt = tele_query();          /* update telescope NEW v0013 */
    double dt = 1.0+exptime-qt;
    if (dt > 0) msleep((int)(1000.0*dt));
    err = zwo_data(dayCam,data,2);
  }
  if (!err) { char file[512]; struct tm res;
    strftime(tst,sizeof(tst),"%Y%m%dut%H%M%S",gmtime_r(&ut,&res));
    if (dayCam->bits == 24) {          /* RGB data */
#if defined(WRITE_PNG)
      /* 0: 60 MB -- 3 seconds */
      /* 1: 24 MB -- 4 seconds */
      /* 4: 21 MB -- 8 seconds */
      sprintf(file,"%s/%s_%s%s.png",datadir,site,"Y",tst);
      if (write_png(file,data,dayCam->aoiW,dayCam->aoiH,1)) {
        sprintf(buf,"failed to open file '%s'",file);
        logit(0,PREFUN,buf,2);  
      } else {
        logit(0,"write PNG",file,2);
        write_meta(file,ut,exptime);
      }
#endif
#if defined(WRITE_JPG) || defined(WRITE_JPEG) /* v0012 */
      /* 100: 10  MB -- 0.52 sec */
      /*  75:  1  MB -- 0.22 sec */
      /*  50: 0.7 MB -- 0.21 sec */
      /*  25: 0.5 MB -- 0.19 sec */
      sprintf(file,"%s/%s_%s%s.jpg",datadir,site,"Y",tst);
      if (write_jpg(file,data,dayCam->aoiW,dayCam->aoiH,75)) {
        sprintf(buf,"failed to open file '%s'",file);
        logit(0,PREFUN,buf,2);  
      } else {
        logit(0,"write JPEG",file,2);
        write_meta(file,ut,exptime);   /* NEW v0013 */
      }
#endif
#if 0 /* PPM --> png/jpeg */
      char tmp[32]="/tmp/casca.ppm",cmd[1024];
      double t1 = walltime(0);
      FILE *fp = fopen(tmp,"wb");
      fprintf(fp,"P6 %d %d 255\n",dayCam->aoiW,dayCam->aoiH);
      fwrite(data,3,npix,fp);
      fclose(fp);
      double t2 = walltime(0);
      sprintf(file,"%s/%s_%s%s.png",datadir,site,"Y",tst);
      sprintf(cmd,"convert %s %s",tmp,file);     /* 21 MB -- 24.9 sec */
      sprintf(cmd,"pnmtopng %s > %s",tmp,file);  /* 43 MB --  4.5 sec */
      sprintf(file,"%s/%s_%s%s.jpeg",datadir,site,"Y",tst);
      sprintf(cmd,"pnmtojpeg %s > %s",tmp,file); /*  1 MB --  0.7 sec */
      system(cmd);
      double t3 = walltime(0);
      printf("PPM/PNG: %s (%.2f,%.2f)\n",cmd,t2-t1,t3-t2);
#endif
    } else {
      (void)zwo_temperature(dayCam);   /* v0012 */
      FITS* f = fits_create(dayCam->aoiW,dayCam->aoiH,dayCam->bits);
      if (dayCam->bits == 8) {  f->bitpix = 8;  f->bzero = 0; }
      sprintf(file,"%s/%s_%s%s.fits",datadir,site,"Y",tst);
      if (fits_open(f,file) == 0) {
        logit(0,"write FITS",file,2);
        fits_char (f,"DATE-OBS",prtime(ut,buf),NULL);
        fits_char (f,"INSTRUME",instrument,NULL);
        fits_float(f,"SITELONG",sitelong,5,NULL);
        fits_float(f,"SITELAT",sitelat ,5,NULL);
        fits_float(f,"SITEELEV",siteelev,0,NULL);
        fits_float(f,"PIXSIZE",pixsize,3,NULL);
        fits_float(f,"EXPTIME",exptime,4,NULL);
        fits_char (f,"WINDOW",window,NULL);
        fits_int  (f,"BINNING",binning,"CCD binning");
        fits_int  (f,"CCD-GAIN",day_gain,"CCD gain"); /* v0012 */
        fits_int  (f,"CCD-OFFS",day_offset,"CCD offset");
        find_object(ut,0,&up,&rise,&set); /* v0013 */
        fits_char (f,"SUNRISE",prtime(rise,buf),"sun rise");
        fits_char (f,"SUNSET",prtime(set,buf),"sun set");
        get_sun(ut,sitelong,sitelat,&ra,&dec,&el,&az);
        fits_float(f,"SUNAZ",az,4,"sun azimuth");
        fits_float(f,"SUNEL",el,4,"sun elevation");
        if (*baade_host || *clay_host) { /* v0013 */
          fits_float(f,"ENVTEM",env_temp,2,"ENV temperature");
          fits_float(f,"ENVPRE",env_pres,2,"ENV pressure");
          fits_float(f,"ENVHUM",env_humi,2,"ENV humidity");
          fits_float(f,"ENVWIN",env_wind,2,"ENV wind speed");
          fits_float(f,"ENVDIR",env_wdir,2,"ENV wind direction");
        }
        fits_char (f,"SOFTWARE",P_VERSION,"software version");
        fits_endHeader(f);
        fits_writeData(f,data);
        fits_close(f);
      } else { char buffer[1024];
        sprintf(buffer,"failed to open file '%s'",file);
        logit(0,PREFUN,buffer,2);  
      }
      fits_free(f); 
    }
  } /* endif(err) -- reported by caller */
  free((void*)data);

  return err;
}


/* ---------------------------------------------------------------- */

static int move_filter(Observation* obs)
{
  char buf[128];

  sprintf(buf,"'%d:%s'",obs->number,obs->name);
  logit(0,PREFUN,buf,1);
  int err = zwo_filter(zwoCam,obs->number,1);
  if (err) { 
    sprintf(buf,"moving filter failed (err=%d)",err);
    logit(0,PREFUN,buf,2);
  }
  return err;
}

/* ---------------------------------------------------------------- */

static char* get_def_string(char* str,const char* base,int i,char* def)
{
  char buf[128];
  sprintf(buf,"%s%d",base,i);
  get_string(configfile,buf,str,"__default__");
  if (!strcmp(str,"__default__")) get_string(configfile,base,str,def);
  return str;
}

/* --- */

static long get_def_long(const char* base,int i,long def)
{
  char buf[128];
  sprintf(buf,"%s%d",base,i);
  long number = get_long(configfile,buf,-99999999);
  if (number == -99999999) number = get_long(configfile,base,def);
  return number;
}

/* --- */

static double get_def_double(const char* base,int i,double def)
{
  char buf[128];
  sprintf(buf,"%s%d",base,i);
  double number = get_double(configfile,buf,-99999999);
  if (number == -99999999) number = get_double(configfile,base,def);
  return number;
}  

/* ---------------------------------------------------------------- */

static void casca_shutdown(ZwoStruct* zwo,const char* reason)
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p,%s)\n",PREFUN,zwo,reason);
#endif
  assert(zwo->sock > 0);

  logit(0,PREFUN,reason,0);            /* close logfile */
  zwo_close(zwo);
  if (zwo == zwoCam) tele_close();
  state = STATE_IDLE;
}

/* ---------------------------------------------------------------- */

int main(int argc,char** argv)
{
  int    i,port=1,err,isDay=0,isNight=0;
  double ra,dec,sun_az,sun_el;
  char buf[512],buffer[1024];
  TCPIP_ServerInfo info;

  /* parse command line ------------------------------------------- */
  sprintf(configfile,"%s/casca.conf",getenv("HOME"));
  { extern char *optarg;
    extern int opterr,optopt,optind; opterr=0;
    while ((i=getopt(argc,argv,"c:o:p:s:")) != EOF) {
      switch (i) {
      case 'c':                        /* config file */
        strcpy(configfile,optarg);
        break;
      case 'o':                        /* offtime */
#ifdef SIM_ONLY
        offtime = -3600*atoi(optarg);
#endif
        break;
      case 'p':
        port = atoi(optarg);
        break;
      case 's':
#ifdef SIM_ONLY
        speedup = atoi(optarg);
#endif
        break;
      case '?':
        fprintf(stderr,"%s: option '-%c' unknown or parameter missing\n",
                P_TITLE,(char)optopt);
        exit(1);
      }
    }
    if (optind < argc) fprintf(stderr,"unknown %s\n",argv[optind]);
  }
  sprintf(buf,"version= %s , LT= %s",P_VERSION,get_local_timestr(buffer,0));
  logit(0,P_TITLE,buf,1);
  sprintf(buf,"config= %s",configfile);
  logit(0,P_TITLE,buf,1);

  /* startup TCP/IP server thread --------------------------------- */
  if (port != 1) info.port = port;
  else info.port = get_long(configfile,"CASCAPORT",50002+100*PROJECT_ID);
  if (info.port) {
    info.name = P_TITLE;
    info.oklist = NULL;
    info.handler = handle_tcpip_request;
    i = TCPIP_StartupServerThread(&info,2000);
    if (i) {
      fprintf(stderr,"%s: creating TCPIP-server thread failed (err=%d)\n",
             P_TITLE,i);
      exit(1); 
    }
#if (DEBUG > 0)
    printf("%s listens on port %d\n",P_TITLE,info.port);
#endif
  }

  /* read configuration file -------------------------------------- */

  get_string(configfile,"DATADIR",datadir,getenv("HOME"));
  get_string(configfile,"SITE",site,"lc");
  sitelong = get_double(configfile,"SITELONG",-70.69284);
  sitelat = get_double(configfile,"SITELAT",-29.01392);
  siteelev = get_double(configfile,"SITEL",2405.0); 

  get_string(configfile,"DETECTOR",detector,"ASI1600MM");
  pixsize = get_double(configfile,"PIXSIZE",3.8);
  ccd_setpoint = get_long(configfile,"CCD-TEMP",-5); /* -15 */
  double ccd_range = get_double(configfile,"CCD-RANGE",0.5);
  ccd_gain = get_long(configfile,"CCD-GAIN",0); /* 138 */
  ccd_offset = get_long(configfile,"CCD-OFFSET",10); /* 32 */
  binning = get_long(configfile,"BINNING",1); /* v0010 */
  get_string(configfile,"WINDOW",window,""); /* [662:4061,88:3487] */
  double night = get_double(configfile,"NIGHT",-9.0);

  get_string(configfile,"ZWOHOST",buf,DEF_ZWOHOST);
  i = get_long(configfile,"ZWOPORT",50011+100*PROJECT_ID); /* 52311 */
  if (*buf && (i > 0)) zwoCam = zwo_create(buf,i);
  sprintf(buffer,"zwoCam=%p (%s:%d)",zwoCam,buf,i);
  logit(0,PREFUN,buffer,1);

  nfilters = get_long(configfile,"NFILTERS",0);
  int nseq = (nfilters > 0) ? nfilters : 1;
  observations = malloc(nseq*sizeof(Observation));
  for (i=0; i<nseq; i++) {
    Observation *obs = &observations[i];
    get_def_string(obs->name,"FILTERNAME",1+i,"X");
    obs->number = get_def_long("FILTERNUMBER",1+i,i);
    obs->time_dn = get_def_double("EXPTIME_DOWN",1+i,0);
    obs->time_up = get_def_double("EXPTIME_UP",1+i,0);
    if (obs->time_dn<=0) obs->time_dn = (obs->time_up<=0) ? 1 : obs->time_up;
    if (obs->time_up<=0) obs->time_up = (obs->time_dn<=0) ? 1 : obs->time_dn;
    sprintf(buf,"%s %d %6.3f %6.3f",obs->name,obs->number,
            obs->time_dn,obs->time_up);
    logit(0,"sequence",buf,1);
  }
//#ifndef SIM_ONLY
  period = get_long(configfile,"PERIOD",60);
//#endif

  /* --- telescope setup ------------------------------------------ */

  get_string(configfile,"BAADEHOST", baade_host, "");
  logit(0,"Baade",baade_host,1);
  get_string(configfile,"CLAYHOST",  clay_host,  "");
  logit(0,"Clay",clay_host,1);
  get_string(configfile,"DIMMHOST",  dimm_host,  "");
  logit(0,"DIMM",dimm_host,1);
  get_string(configfile,"DUPONTHOST",dupont_host,"");
  logit(0,"DuPont",dupont_host,2);

  /* --- daytime (color) camera ----------------------------------- */

  get_string(configfile,"DAYHOST",buf,DEF_DAYHOST);
  i = get_long(configfile,"DAYPORT",50011+100*PROJECT_ID); /* 52311 */
  if (*buf && (i > 0)) dayCam = zwo_create(buf,i);
  sprintf(buffer,"dayCam=%p (%s:%d)",dayCam,buf,i);
  logit(0,PREFUN,buffer,1);

  day_gain = get_long(configfile,"DAY-GAIN",0); /* v0012 */
  day_offset = get_long(configfile,"DAY-OFFSET",10); 
  day_exptime = get_double(configfile,"DAY-EXPTIME",0.1); 

  /* --- main loop ------------------------------------------------ */

  while (!done) {
    sleep(1);
    if (suspend) continue; // todo what if observing? -- shutdown ?
#ifdef SIM_ONLY
    if (speedup) { offtime -= sim_delta; sim_delta += speedup; }
#endif
    time_t ut = cor_time(offtime);
    if (ut-last_loop < period) continue;
    last_loop = ut;                    /* NEW v0013 */
    get_sun(ut,sitelong,sitelat,&ra,&dec,&sun_el,&sun_az);
    switch (state) {                   /* state machine */
    case STATE_IDLE:
#if (DEBUG > 0)
      fprintf(stderr,"%s sunEl=%f\n",prtime(ut,buf),sun_el);
#endif
      if (sun_el < night) {            /* beginning of night */
        isDay = 0;
        if (zwoCam) {   
          if (!isNight) logit(ut,PREFUN,"beginning of night",2);
          isNight = 1;
          err = setup_camera(zwoCam,window,16);
          if (!err) {
            if (nfilters==1) err = move_filter(&observations[0]);
            if (err) casca_shutdown(zwoCam,"filter wheel failed");
          }
          if (err) {
            logit(0,PREFUN,"CASCA night startup failed",2);
          } else {
            state = STATE_NIGHT; last_loop = 0;  /* ASAP */
          }
        }
      } else 
      if ((sun_el > 0) && (sun_el > night)) { /* beginnig of daytime */
        isNight = 0;
        if (dayCam) {
          if (!isDay) logit(ut,PREFUN,"beginning of day",2);
          isDay = 1;
#if defined(WRITE_JPEG) || defined(WRITE_JPG) || defined(WRITE_PNG)
          err = setup_camera(dayCam,"",24);
#else
          err = setup_camera(dayCam,"",16);
#endif
          if (err) {
            logit(0,PREFUN,"CASCA day startup failed",2);
          } else {
            state = STATE_DAY; last_loop = 0; /* ASAP */
          }
        }
      } else {
        isDay = isNight = 0; 
      }
      break;
    case STATE_NIGHT:                  /* what happens at night */
      assert(zwoCam); assert(zwoCam->sock > 0);
      if (sun_el > night) {            /* end of night */
        casca_shutdown(zwoCam,"end-of-night");
        break;
      }
      if (0 == zwo_temperature(zwoCam)) { 
        if (zwoCam->hasCooler) {       /* read CCD temperature v0010 */
          if (zwoCam->tempSensor-ccd_setpoint > ccd_range) {
            sprintf(buf,"CCDtemp= %.1f Cooler= %.0f CCDsetp= %d", // v0011
                    zwoCam->tempSensor,zwoCam->percent,ccd_setpoint);
            logit(0,PREFUN,buf,2);
            break;
          }
        }
      }
      sprintf(buf,"exposure loop (%d) sunEl= %.1f CCDtemp= %.1f Cooler= %.0f",
              nseq,sun_el,zwoCam->tempSensor,zwoCam->percent); // v0011
      logit(ut,PREFUN,buf,2);
      assert(observations);
      for (i=0,err=0; i<nseq; i++) {
        Observation *obs = &observations[i];
        if (nfilters > 1) err = move_filter(obs);
        if (!err) err = take_nightime_exposure(obs);
        if (err) break;
        if (done) break; // ?Povilas finish sequence
      }
      if (err) {
        sprintf(buf,"failed night time exposure, err=%d",err);
        casca_shutdown(zwoCam,buf);
      }
      break;
    case STATE_DAY:
      assert(dayCam); assert(dayCam->sock > 0);
      if ((sun_el <= 0) || (sun_el < night)) { /* end of day */
        casca_shutdown(dayCam,"end-of-day");
        break;
      }
#ifndef SIM_ONLY
      assert(dayCam->isColor);
#endif
      sprintf(buf,"take color exposure (sunEl=%.1f)",sun_el);
      logit(0,PREFUN,buf,2);
      err = take_daytime_exposure(ut);
      if (err) {
        sprintf(buf,"failed daytime exposure, err=%d",err);
        casca_shutdown(dayCam,buf);
      }
      break;
    } /* end-switch (state) */
#ifdef SIM_ONLY
    if (state == STATE_IDLE) sim_delta=0;
#endif
  }
  logit(0,PREFUN,"terminate",0);

  if (zwoCam) { zwo_free(zwoCam); zwoCam = NULL; } /* implicit close */
  if (dayCam) { zwo_free(dayCam); dayCam = NULL; }
  tele_close();

  if (info.port) {                     /* terminate TCP/IP server */
    (void)TCPIP_TerminateServerThread(&info,200);
  }

  return 0;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
